#define LGFX_M5PAPER
#define LGFX_USE_V1
#define LGFX_AUTODETECT

#include "keypad.hpp"
#include <M5EPD.h>
#include <LovyanGFX.hpp>

static LGFX gfx;

constexpr float FONT_SIZE_NORMAL = 4.0;
constexpr float FONT_SIZE_LARGE = 8.0;

static const float_t BUTTON_WIDTH = 115;
static const float_t BUTTON_HEIGHT = 115;
static int last_index = -1;

static String oldinput = "";
static String input = "";

int index(int x, int y)
{
    for (int index = 0; index <= 11; index++)
    {
        float_t cx = (125 * (index % 3)) + 15;
        float_t cy = (125 * (index / 3)) + 20;

        if (x > cx && x < cx + BUTTON_WIDTH && y > cy && y < cy + BUTTON_HEIGHT)
        {
            return index;
        }
    }

    return -1;
}

void drawButton(int index, String label)
{
    float_t x = (125 * (index % 3)) + 15;
    float_t y = (125 * (index / 3)) + 20;

    gfx.setCursor(x + (115 / 2) - (gfx.fontWidth() / 2), y + (115 / 2) - (gfx.fontHeight() / 2));
    gfx.print(label);

    gfx.setColor(TFT_BLACK);
    gfx.drawFastHLine(x, y, BUTTON_WIDTH);
    gfx.drawFastHLine(x, y + BUTTON_HEIGHT, BUTTON_WIDTH);
    gfx.drawFastVLine(x, y, BUTTON_HEIGHT);
    gfx.drawFastVLine(x + BUTTON_WIDTH, y, BUTTON_HEIGHT);
}

void drawButtons()
{
    gfx.setTextSize(FONT_SIZE_NORMAL);
    for (int i = 0; i <= 11; i++)
    {
        String label = String(i + 1);
        if (i == 9)
        {
            label = String("<");
        }
        if (i == 10)
        {
            label = String("0");
        }
        if (i == 11)
        {
            label = String(">");
        }

        drawButton(i, label);
    }
}

void drawInputField()
{
    int x = 405;
    int y = 20;

    gfx.setColor(TFT_BLACK);
    gfx.drawFastHLine(x, y, 520);
    gfx.drawFastHLine(x, y + 115, 520);
    gfx.drawFastVLine(x, y, 115);
    gfx.drawFastVLine(x + 520, y, 115);
}

void drawInput()
{
    int x = 405;
    int y = 20;

    gfx.setEpdMode(epd_mode_t::epd_text);
    gfx.setColor(TFT_BLACK);
    gfx.setTextSize(FONT_SIZE_LARGE);

    if (input.length() < oldinput.length())
    {
        gfx.setColor(TFT_WHITE);
        gfx.fillRect(x + 1, y + 1, 519, 114);
        gfx.setColor(TFT_BLACK);
    }

    gfx.setCursor(x + (520 / 2) - ((gfx.fontWidth() * input.length()) / 2), y + (115 / 2) - (gfx.fontHeight() / 2));
    String keypad = "";
    for (int i = 0; i < input.length(); i++)
    {
        keypad += "*";
    }
    gfx.print(keypad);
    gfx.setEpdMode(epd_mode_t::epd_fast);
}

void keypad_init()
{
    gfx.init();
    gfx.setRotation(1);

    gfx.setEpdMode(epd_mode_t::epd_quality);
    gfx.fillScreen(TFT_WHITE);
    gfx.waitDisplay();
    gfx.fillScreen(TFT_WHITE);
    gfx.setEpdMode(epd_mode_t::epd_fast);

    gfx.setColor(TFT_BLACK);
    gfx.waitDisplay();

    drawButtons();
    drawInputField();
    drawInput();
}

void keypad_write(String payload)
{
    payload.replace("_", " ");

    gfx.setTextSize(FONT_SIZE_NORMAL);

    gfx.setColor(TFT_WHITE);
    gfx.fillRect(405, 155, gfx.fontWidth() * 10, gfx.fontHeight());
    gfx.setColor(TFT_BLACK);

    gfx.setCursor(405, 155);
    payload[0] = toupper(payload[0]);
    gfx.printf("%s", payload);
}

std::optional<String> checkForInput()
{
    if (M5.TP.available())
    {
        M5.TP.update();
        if (M5.TP.isFingerUp())
        {
            last_index = -1;
            return std::nullopt;
        }

        tp_finger_t fingerItem = M5.TP.readFinger(0);

        int i = index(fingerItem.x, fingerItem.y);
        if (last_index != i)
        {
            last_index = i;
            if (i == 9)
            {
                // revert
                input = "";
            }
            else if (i == 10)
            {
                if (input.length() >= 6)
                    return std::nullopt;
                input += String(0);
            }
            else if (i == 11)
            {
                // submit
                String retval = input;
                input = "";
                return retval;
            }
            else
            {
                if (input.length() >= 6)
                    return std::nullopt;
                input += String(i + 1);
            }
        }
    }
    return std::nullopt;
}

std::optional<String> keypad_loop()
{
    auto value = checkForInput();
    if (oldinput != input)
    {
        drawInput();
        oldinput = input;
    }

    return value;
}